import time
import numpy as np
import scipy as sp
import inu
import itrm


def fake_path(T, axis=1):
    # Define time.
    K = round(360.0/T) + 1
    t = np.arange(K)*T

    # Define a figure eight.
    R = 0.000156784 # radians
    theta = np.linspace(0, 2*np.pi, K)
    lat = (R/4)*np.sin(2*theta)
    lon = R*(np.cos(theta) - 1)
    hae = 50.0*(1 - np.cos(theta))
    llh_t = np.row_stack((lat, lon, hae))

    # Transpose.
    if axis == 0:
        llh_t = llh_t.T

    return t, llh_t


def view_jacobian():
    # Build path.
    T = 0.1
    t, llh_t = fake_path(T)
    K = len(t)
    hae_t = llh_t[2]
    vne_t = inu.llh_to_vne(llh_t, T)
    grav_t = inu.somigliana(llh_t)
    rpy_t = inu.vne_to_rpy(vne_t, grav_t[2, :], T)

    # Inverse mechanize.
    fbbi_t, wbbi_t = inu.inv_mech(llh_t, rpy_t, T)

    # -------------------------------
    # Define the small delta vectors.
    # -------------------------------

    # position
    dang = 1e-7
    dh = 1e-1
    dlat = np.array([dang, 0, 0])
    dlon = np.array([0, dang, 0])
    dhae = np.array([0, 0, dh])

    # velocity
    dv = 1e-1
    dvN = np.array([dv, 0, 0])
    dvE = np.array([0, dv, 0])
    dvD = np.array([0, 0, dv])

    # tilt error
    psi = 1e-6
    psix = np.array([psi, 0, 0])
    psiy = np.array([0, psi, 0])
    psiz = np.array([0, 0, psi])
    Psix = np.array([
        [0, -psix[2], psix[1]],
        [psix[2], 0, -psix[0]],
        [-psix[1], psix[0], 0]])
    Psiy = np.array([
        [0, -psiy[2], psiy[1]],
        [psiy[2], 0, -psiy[0]],
        [-psiy[1], psiy[0], 0]])
    Psiz = np.array([
        [0, -psiz[2], psiz[1]],
        [psiz[2], 0, -psiz[0]],
        [-psiz[1], psiz[0], 0]])
    Delx = sp.linalg.expm(Psix.T)
    Dely = sp.linalg.expm(Psiy.T)
    Delz = sp.linalg.expm(Psiz.T)

    # ---------------------------------------
    # Initialize states and allocate storage.
    # ---------------------------------------

    # Initialize the states.
    llh = llh_t[:, 0]
    vne = vne_t[:, 0]
    Cnb = inu.rpy_to_dcm(rpy_t[:, 0]).T

    # Allocate storage.
    F_t = np.zeros((K, 9, 9))
    Fn_t = np.zeros((K, 9, 9))
    F60 = np.zeros(K)

    # -----------------
    # Process the data.
    # -----------------

    tic = time.perf_counter()
    for k in range(K):
        # Inputs
        fbbi = fbbi_t[:, k]
        wbbi = wbbi_t[:, k]

        # Override height and velocity for stability.
        llh[2] = hae_t[k]
        if k < K - 1:
            vne[2] = -(hae_t[k + 1] - hae_t[k])/T

        # Get the Jacobian.
        F_t[k, :, :] = inu.jacobian(fbbi, llh, vne, Cnb)

        # Get correct derivatives.
        Dllh, Dvne, wbbn = inu.mech_step(fbbi, wbbi, llh, vne, Cnb)
        E = inu.rodrigues_rotation(wbbn*T) # exp([w]x T)
        Dx = np.concatenate((Dllh, Dvne, np.zeros(3)))

        # Get partial of states with respect to position.
        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh + dlat, vne, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 0] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dang

        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh + dlon, vne, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 1] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dang

        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh + dhae, vne, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 2] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dh

        # Get partial of states with respect to velocity.
        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne + dvN, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 3] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dv

        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne + dvE, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 4] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dv

        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne + dvD, Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 5] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/dv

        # Get partial of states with respect to tilt errors.
        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne, Delx @ Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 6] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/psi

        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne, Dely @ Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 7] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/psi
        
        tDllh, tDvne, twbbn = inu.mech_step(fbbi, wbbi, llh, vne, Delz @ Cnb)
        tE = inu.rodrigues_rotation(twbbn*T) # exp([w~]x T)
        tDpsi = -inu.inverse_rodrigues_rotation(Cnb @ tE @ E.T @ Cnb.T)/T
        Fn_t[k, :, 8] = (np.concatenate((tDllh, tDvne, tDpsi)) - Dx)/psi

        # Integrate.
        llh += Dllh * T
        vne += Dvne * T
        Cnb[:, :] = Cnb @ inu.rodrigues_rotation(wbbn * T)
        inu.orthonormalize_dcm(Cnb)

        # Update progress bar.
        inu.progress(k, K, tic)

    # Build time.
    t = np.arange(K)*T

    # Plot the comparison of the results.
    names = ["lat", "lon", "hae", "vN", "vE", "vD", "ψx", "ψy", "ψz"]
    for row in range(9):
        for col in range(9):
            F = F_t[:, row, col]
            Fn = Fn_t[:, row, col]
            if (np.max(F) < 1e-16) and (np.min(F) > -1e-16):
                continue

            # Get the NMAE.
            er = F - Fn
            rmse = np.sqrt(np.mean(er**2))

            itrm.iplot(t, np.array([Fn, F]),
                       label=f"{rmse*100:.4g}, d{names[row]}/d{names[col]}",
                    rows=0.333)

view_jacobian()



import time
import numpy as np
import matplotlib.pyplot as plt
import inu
import r3f
import itrm


def get_path(T):
    # Define time.
    t_dur = 3600.0
    K = round(t_dur/T) + 1
    t = np.arange(K)*T

    # Define a figure eight.
    R = 0.016 # lat-long radius (rad)
    theta = np.linspace(0, 2*np.pi, K)
    lat = (R/4)*np.sin(2*theta)
    lon = R*(np.cos(theta) - 1)
    a = 0.06
    hae = 50.0*(1 + a - np.cos(theta) - a*np.cos(20*theta))
    llh_t = np.array((lat, lon, hae))

    vne_t = inu.llh_to_vne(llh_t, T)
    grav_t = inu.somigliana(llh_t)
    rpy_t = inu.vne_to_rpy(vne_t, grav_t[2, :], T)

    return t, llh_t, vne_t, rpy_t


def est(fbbi_t, wbbi_t, z_t, # specific forces, rotation rates, GPS positions
        llh, vne, Cnb, # initial states
        K, T): # samples and sampling period

    # Define the constant matrices.
    Qd = np.diag([ # discretized dynamics noise matrix
            4e-6, 4e-6, 100.0, # latitude, longitude, height
            50, 50, 50, # vN, vE, vD
            1e-4, 1e-4, 1e-4])**2 * T # tilt angles
    I = np.eye(9) # identity matrix
    R = np.diag([1.5e-5, 1.5e-5, 200.0])**2 # GPS covariance matrix
    H = np.zeros((3, 9)) # measurement matrix
    H[:3, :3] = np.eye(3)

    # Initialize the state covariance matrix.
    Ph = np.eye(9) # "P hat"

    # Allocate storage.
    hllh_t = np.zeros((3, K))
    hvne_t = np.zeros((3, K))
    hrpy_t = np.zeros((3, K))

    tic = time.perf_counter()
    for k in range(K):
        # Inputs
        fbbi = fbbi_t[:, k] # specific forces (m/s^2)
        wbbi = wbbi_t[:, k] # rotation rates (rad/s)
        z = z_t[:, k] # GPS position (rad, rad, m)

        # Update
        S = H @ Ph @ H.T + R # innovation covariance (3, 3)
        Si = np.linalg.inv(S) # inverse (3, 3)
        Kg = Ph @ H.T @ Si # Kalman gain (9, 3)
        Ph -= Kg @ H @ Ph # update to state covariance (9, 9)
        r = z - llh # innovation (3,)
        dx = Kg @ r # changes to states (9,)
        llh += dx[:3] # add change in position
        vne += dx[3:6] # add change in velocity
        # matrix exponential of skew-symmetric matrix
        Psi = inu.rodrigues_rotation(dx[6:])
        Cnb = Psi.T @ Cnb

        # Save results.
        hllh_t[:, k] = llh
        hvne_t[:, k] = vne
        hrpy_t[:, k] = inu.dcm_to_rpy(Cnb.T)

        # Get the Jacobian and propagate the state covariance.
        F = inu.jacobian(fbbi, llh, vne, Cnb)
        Phi = I + (F*T)@(I + (F*T/2)) # 2nd-order expm(F T)
        Ph = Phi @ Ph @ Phi.T + Qd

        # Get the state derivatives.
        Dllh, Dvne, wbbn = inu.mech_step(fbbi, wbbi, llh, vne, Cnb)

        # Integrate (forward Euler).
        llh += Dllh * T # change applies linearly
        vne += Dvne * T # change applies linearly
        Cnb[:, :] = Cnb @ inu.rodrigues_rotation(wbbn * T)
        inu.orthonormalize_dcm(Cnb)

        # Update progress bar.
        inu.progress(k, K, tic)

    return hllh_t, hvne_t, hrpy_t


# Constants
T = 0.1
Qd = np.diag([ # discretized dynamics noise matrix
        4e-6, 4e-6, 100.0, # latitude, longitude, height
        50, 50, 50, # vN, vE, vD
        1e-5, 1e-5, 1e-5])**2 * T # tilt angles
R = np.diag([1.5e-5, 1.5e-5, 200.0])**2 # GPS covariance matrix
qf = (1e-4)*T*np.eye(3) # accelerometer noise covariance
qw = (1e-6)*T*np.eye(3) # gyroscope noise covariance

# Get a fake path.
t, llh_t, vne_t, rpy_t = get_path(T)

# Get the "GPS" measurements.
K = llh_t.shape[1]
z_t = llh_t + np.linalg.cholesky(R) @ np.random.randn(3, K)

# Inverse mechanize paths to get sensor values and add noise.
print("Inverse mechanization ...")
fbbi_t, wbbi_t = inu.inv_mech(llh_t, vne_t, rpy_t, T)
fbbi_t += np.linalg.cholesky(qf) @ np.random.randn(3, K)
wbbi_t += np.linalg.cholesky(qw) @ np.random.randn(3, K)

# Run the Extended Kalman Filter (EKF).
print("Extended Kalman Filter ...")
llh = llh_t[:, 0].copy()
vne = vne_t[:, 0].copy()
Cnb = inu.rpy_to_dcm(rpy_t[:, 0].copy()).T
hllh_t, hvne_t, hrpy_t = est(fbbi_t, wbbi_t, z_t, llh, vne, Cnb, K, T)

# Convert the estimate and the truth to curvilinear coordinates.
[xc, yc, zc] = r3f.geodetic_to_curvilinear(llh_t)
[hxc, hyc, hzc] = r3f.geodetic_to_curvilinear(hllh_t)

# Show the comparison of true path and estimate.
itrm.plot([yc/1e3, hyc/1e3], [xc/1e3, hxc/1e3],
    equal_axes=True, label="true and est")

# Show the errors.
er = np.array([xc - hxc, yc - hyc, zc - hzc])
itrm.iplot(t/60, er, label="curvilinear errors (m)")

print(f"STD:", np.std(er, axis=1))

er = rpy_t - hrpy_t
itrm.iplot(t/60, er, label="attitude errors (rad)")

itrm.iplot(t/60, [rpy_t[0], hrpy_t[0]], label="true and est roll (rad)")
itrm.iplot(t/60, [rpy_t[1], hrpy_t[1]], label="true and est pitch (rad)")
itrm.iplot(t/60, [rpy_t[2], hrpy_t[2]], label="true and est yaw (rad)")



"""
This script demonstrates the inherent instability of trapezoidal
differentiation. If the averaging weight, alpha, is exactly 0.5, there will be a
bounded oscillation. If it is less than 0.5, the oscillation will grow with
time.
"""

import numpy as np
import itrm


def dtrapz(x, T):
    alpha = 0.5
    K = len(x)
    x_ext = np.append(x, 3*x[-1] - 3*x[-2] + x[-3])
    Dx = (x_ext[1] - x_ext[0])/T
    Dx_t = np.zeros(K)
    for k in range(K):
        Dx_t[k] = Dx
        Dx = (x_ext[k+1] - x_ext[k])/(T*alpha) - (1 - alpha)/alpha*Dx
    return Dx_t


# Create a time array and sine wave.
t = np.linspace(0, 1, 100000)
T = t[1] - t[0]
w = 2*np.pi*10
x = np.sin(w*t)

# Calculate the true analytical derivative.
dx = w*np.cos(w*t)

# Calculate the numerical derivative.
Dx = dtrapz(x, T)

# Plot the difference.
itrm.iplot(t, (dx - Dx)/(2*w), label="normalized derivative error")
